@@ -7,6 +7,7 @@ from __future__ import with_statement
 
 import asyncio
 import logging
+from collections import defaultdict
 from datetime import datetime
 from typing import (
     Any,
@@ -23,12 +24,11 @@ from slixmpp import (
     JID,
 )
 from slixmpp.plugins import BasePlugin
-from slixmpp.xmlstream import register_stanza_plugin, ET
+from slixmpp.xmlstream import register_stanza_plugin
 from slixmpp.xmlstream.handler.callback import Callback
-from slixmpp.xmlstream.matcher.xpath import MatchXPath
 from slixmpp.xmlstream.matcher.stanzapath import StanzaPath
 from slixmpp.xmlstream.matcher.xmlmask import MatchXMLMask
-from slixmpp.exceptions import IqError, IqTimeout, PresenceError
+from slixmpp.exceptions import PresenceError
 
 from slixmpp.plugins.xep_0004 import Form
 from slixmpp.plugins.xep_0045 import stanza
@@ -69,20 +69,33 @@ ROLES = ('moderator', 'participant', 'visitor', 'none')
 class XEP_0045(BasePlugin):
 
     """
-    XEP-0045 Multi-User Chat
+    XEP-0045 Multi-User Chat.
+
+
+    This XEP is made for use by *clients* or components acting as clients.
+    If a component is made to act as MUC service, it should only care about
+    the events and the stanza plugins.
+
+    A single component may want to simulate joins as multiple entities, in
+    which case it should enable the multi_from config option.
+
+
     """
 
     name = 'xep_0045'
     description = 'XEP-0045: Multi-User Chat'
-    dependencies = {'xep_0030', 'xep_0004', 'xep_0203'}
+    dependencies = {'xep_0004', 'xep_0030', 'xep_0172', 'xep_0203'}
     stanza = stanza
+    default_config = {
+        'multi_from': False,
+    }
 
-    rooms: Dict[JID, Dict[str, MucRoomItem]]
-    our_nicks: Dict[JID, str]
+    rooms: Dict[Optional[JID], Dict[JID, Dict[str, MucRoomItem]]]
+    our_nicks: Dict[Optional[JID], Dict[JID, str]]
 
     def plugin_init(self):
-        self.rooms = {}
-        self.our_nicks = {}
+        self.rooms = defaultdict(lambda: defaultdict())
+        self.our_nicks = defaultdict(lambda: defaultdict())
         # load MUC support in presence stanzas
         register_stanza_plugin(MUCMessage, MUCUserItem)
         register_stanza_plugin(MUCPresence, MUCUserItem)
@@ -163,6 +176,8 @@ class XEP_0045(BasePlugin):
                 StanzaPath('message/muc/decline'),
                 self._handle_groupchat_decline
         ))
+        if not self.xmpp.is_component:
+            self.multi_from = False
 
     def plugin_end(self):
         self.xmpp.plugin['xep_0030'].del_feature(feature=stanza.NS)
@@ -175,7 +190,7 @@ class XEP_0045(BasePlugin):
         if self.xmpp.is_component:
             self.xmpp.event('groupchat_invite', inv)
         else:
-            if inv['from'] not in self.rooms.keys():
+            if inv['from'] not in self.rooms[None].keys():
                 self.xmpp.event("groupchat_invite", inv)
 
     def _handle_groupchat_decline(self, decl: Message):
@@ -191,28 +206,40 @@ class XEP_0045(BasePlugin):
         self.xmpp.event('groupchat_config_status', msg)
         self.xmpp.event('muc::%s::config_status' % msg['from'].bare, msg)
 
-    def _client_handle_presence(self, pr: Presence):
+    def _handle_presence(self, pr: Presence):
         """As a client, handle a presence stanza"""
         got_offline = False
         got_online = False
-        if pr['muc']['room'] not in self.rooms.keys():
+        if self.multi_from:
+            rooms = self.rooms[pr['to']]
+        else:
+            rooms = self.rooms[None]
+        if pr['muc']['room'] not in rooms.keys():
             return
         self.xmpp.roster[pr['from']].ignore_updates = True
         entry = pr['muc'].get_stanza_values()
         entry['show'] = pr['show'] if pr['show'] in pr.showtypes else None
         entry['status'] = pr['status']
-        entry['alt_nick'] = pr['nick']
+
+        alt_nick_element = pr.get_plugin('nick', check=True)
+        entry['alt_nick'] = alt_nick_element['nick'] if alt_nick_element else ''
         if pr['type'] == 'unavailable':
-            if entry['nick'] in self.rooms[entry['room']]:
-                del self.rooms[entry['room']][entry['nick']]
+            if entry['nick'] in rooms[entry['room']]:
+                del rooms[entry['room']][entry['nick']]
             got_offline = True
         else:
-            if entry['nick'] not in self.rooms[entry['room']]:
+            if entry['nick'] not in rooms[entry['room']]:
                 got_online = True
-            self.rooms[entry['room']][entry['nick']] = entry
-        log.debug("MUC presence from %s/%s : %s", entry['room'],entry['nick'], entry)
-        self.xmpp.event("groupchat_presence", pr)
-        if 110 in pr['muc']['status_codes']:
+            rooms[entry['room']][entry['nick']] = entry
+        log.debug("MUC presence from %s/%s : %s", entry['room'], entry['nick'], entry)
+        status_codes = pr['muc']['status_codes']
+        if 110 in status_codes:
+            if 303 in status_codes:
+                if self.multi_from:
+                    pto = pr['to']
+                else:
+                    pto = None
+                self.our_nicks[pto][pr['from'].bare] = pr['muc']['item_nick']
             self.xmpp.event("muc::%s::self-presence" % entry['room'], pr)
         self.xmpp.event("muc::%s::presence" % entry['room'], pr)
         if got_offline:
@@ -226,10 +253,8 @@ class XEP_0045(BasePlugin):
 
     def _handle_groupchat_presence(self, pr: Presence):
         """ Handle a presence in a muc."""
-        if self.xmpp.is_component:
-            self.xmpp.event('groupchat_presence', pr)
-        else:
-            self._client_handle_presence(pr)
+        self.xmpp.event("groupchat_presence", pr)
+        self._handle_presence(pr)
 
     def _handle_groupchat_join(self, pr: Presence):
         """Received a join presence (as a component)"""
@@ -247,7 +272,6 @@ class XEP_0045(BasePlugin):
         self.xmpp.event('groupchat_message_error', msg)
         self.xmpp.event("muc::%s::message_error" % msg['from'].bare, msg)
 
-
     def _handle_groupchat_subject(self, msg: Message):
         """ Handle a message coming from a muc indicating
         a change of subject (or announcing it when joining the room)
@@ -259,6 +283,54 @@ class XEP_0045(BasePlugin):
         self.xmpp.event('groupchat_subject', msg)
         self.xmpp.event('muc::%s::groupchat_subject' % msg['from'].bare, msg)
 
+    def make_join_stanza(self, room: JID, nick: str, *,
+                         password: Optional[str] = None,
+                         maxchars: Optional[int] = None,
+                         maxstanzas: Optional[int] = None,
+                         seconds: Optional[int] = None,
+                         since: Optional[datetime] = None,
+                         presence_options: Optional[PresenceArgs] = None) -> Presence:
+        """
+        Build the stanza for the MUC join, without sending it.
+
+        Only one of {maxchars, maxstanzas, seconds, since} will be used, in
+        that order.
+
+        .. versionadded:: 1.9.0
+
+        :param room: Room JID
+        :param nick: User nickname in the room
+        :param password: The optional room password.
+        :param maxchars: Max number of characters to return from history.
+        :param maxstanzas: Max number of stanzas to return from history.
+        :param seconds: Fetch history until that many seconds in the past.
+        :param since: Fetch history since that timestamp.
+        :return: The presence stanza to send
+        """
+        if presence_options is None:
+            presence_options = {}
+        elif presence_options.get('type') == 'unavailable':
+            del presence_options['type']
+        pto = JID(room)
+        pto.resource = nick
+        stanza = self.xmpp.make_presence(
+            pto=pto,
+            **presence_options
+        )
+        stanza.enable('muc_join')
+        if password is not None:
+            stanza['muc_join']['password'] = password
+        if maxchars is not None:
+            stanza['muc_join']['history']['maxchars'] = str(maxchars)
+        elif maxstanzas is not None:
+            stanza['muc_join']['history']['maxstanzas'] = str(maxstanzas)
+        elif seconds is not None:
+            stanza['muc_join']['history']['seconds'] = str(seconds)
+        elif since is not None:
+            fmt = self.xmpp.plugin['xep_0082'].format_datetime(since)
+            stanza['muc_join']['history']['since'] = fmt
+        return stanza
+
     async def join_muc_wait(self, room: JID, nick: str, *,
                             password: Optional[str] = None,
                             maxchars: Optional[int] = None,
@@ -275,6 +347,8 @@ class XEP_0045(BasePlugin):
 
         .. versionadded:: 1.8.0
 
+        :param room: Room JID
+        :param nick: User nickname in the room
         :param password: The optional room password.
         :param maxchars: Max number of characters to return from history.
         :param maxstanzas: Max number of stanzas to return from history.
@@ -289,30 +363,28 @@ class XEP_0045(BasePlugin):
         :return: A tuple containing our own presence, the subject, a list
                  of occupants and a list of history messages.
         """
-        if presence_options is None:
-            presence_options = {}
-        stanza = self.xmpp.make_presence(
-            pto="%s/%s" % (room, nick),
-            **presence_options
-        )
-        stanza.enable('muc_join')
-        if password is not None:
-            stanza['muc_join']['password'] = password
-        if maxchars is not None:
-            stanza['muc_join']['history']['maxchars'] = str(maxchars)
-        elif maxstanzas is not None:
-            stanza['muc_join']['history']['maxstanzas'] = str(maxstanzas)
-        elif seconds is not None:
-            stanza['muc_join']['history']['seconds'] = str(seconds)
-        elif since is not None:
-            fmt = self.xmpp.plugin['xep_0082'].format_datetime(since)
-            stanza['muc_join']['history']['since'] = fmt
-        self.rooms[room] = {}
-        self.our_nicks[room] = nick
-        stanza.send()
-        return await self._await_join(room, timeout)
+        if self.xmpp.is_component and not presence_options.get('pfrom'):
+            raise ValueError('Components must always set the pfrom= attribute.')
 
-    async def _await_join(self, room: JID, timeout: int = 300) -> JoinResult:
+        stanza = self.make_join_stanza(
+            room=room,
+            nick=nick,
+            password=password,
+            maxchars=maxchars,
+            maxstanzas=maxstanzas,
+            seconds=seconds,
+            since=since,
+            presence_options=presence_options,
+        )
+        if self.multi_from:
+            pfrom = presence_options['pfrom']
+        else:
+            pfrom = None
+        self.rooms[pfrom][room] = {}
+        self.our_nicks[pfrom][room] = nick
+        return await self._await_join(room, stanza, timeout)
+
+    async def _await_join(self, room: JID, stanza: Presence, timeout: int = 300) -> JoinResult:
         """Do the heavy lifting for awaiting a MUC join
 
         A muc join, once the join stanza is sent, is:
@@ -323,22 +395,56 @@ class XEP_0045(BasePlugin):
         history_buffer: List[Message] = []
         occupant_buffer: List[Presence] = []
 
+        pfrom = stanza['from'] or None
+
         def add_message(msg: Message):
+            if pfrom and pfrom != msg['to']:
+                return
             delay = msg.get_plugin('delay', check=True)
             if delay is not None and delay['from'] == room:
                 history_buffer.append(msg)
 
         def add_occupant(pres: Presence):
+            if pfrom and pfrom != pres['to']:
+                return
             occupant_buffer.append(pres)
 
-        catch_occupants = self.xmpp.event_handler("muc::%s::got_online" % room, add_occupant)
-        catch_history = self.xmpp.event_handler("muc::%s::message" % room, add_message)
-        subject_handler = self.xmpp.event_handler("muc::%s::groupchat_subject" % room, topic_received.set_result)
-        self_presence = self.xmpp.event_handler("muc::%s::self-presence" % room, presence_done.set_result)
-        presence_error = self.xmpp.event_handler("muc::%s::presence-error" % room, presence_done.set_result)
+        def set_topic(msg: Message):
+            if pfrom and pfrom != msg['to']:
+                return
+            if not topic_received.done():
+                topic_received.set_result(msg)
+
+        def set_self_presence(pres: Presence):
+            if pfrom and pfrom != pres['to']:
+                return
+            if not presence_done.done():
+                presence_done.set_result(pres)
+
+        catch_occupants = self.xmpp.event_handler(
+            "muc::%s::got_online" % room,
+            add_occupant
+        )
+        catch_history = self.xmpp.event_handler(
+            "muc::%s::message" % room,
+            add_message
+        )
+        subject_handler = self.xmpp.event_handler(
+            "muc::%s::groupchat_subject" % room,
+            set_topic
+        )
+        self_presence = self.xmpp.event_handler(
+            "muc::%s::self-presence" % room,
+            set_self_presence
+        )
+        presence_error = self.xmpp.event_handler(
+            "muc::%s::presence-error" % room,
+            set_self_presence
+        )
 
         with subject_handler, catch_history, catch_occupants:
             with self_presence, presence_error:
+                stanza.send()
                 done, pending = await asyncio.wait(
                     [presence_done],
                     timeout=timeout,
@@ -356,11 +462,13 @@ class XEP_0045(BasePlugin):
                 raise asyncio.TimeoutError()
         subject: Message = topic_received.result()
         # update known nick in case it has changed
-        self.our_nicks[room] = pres['from'].resource
+        self.our_nicks[pfrom][room] = pres['from'].resource
         return (pres, subject, occupant_buffer, history_buffer)
 
-    def join_muc(self, room: JID, nick: str, maxhistory="0", password='',
-                 pstatus='', pshow: PresenceShows='chat', pfrom: JidStr='') -> asyncio.Future:
+    def join_muc(self, room: JID, nick: str, maxhistory: str = "0",
+                 password: str = '', pstatus: str = '',
+                 pshow: PresenceShows = 'chat', pfrom: JidStr = ''
+                 ) -> asyncio.Future:
         """ Join the specified room, requesting 'maxhistory' lines of history.
 
         .. deprecated:: 1.8.0
@@ -398,6 +506,8 @@ class XEP_0045(BasePlugin):
         :param room: Room to leave.
         :param nick: Your nickname.
         :param msg: Presence status to use.
+
+        :raises: KeyError if the room is not in our room list.
         """
         if msg:
             self.xmpp.send_presence(
@@ -412,7 +522,13 @@ class XEP_0045(BasePlugin):
                 pto="%s/%s" % (room, nick),
                 pfrom=pfrom
             )
-        del self.rooms[room]
+        if pfrom in self.rooms and room in self.rooms[pfrom]:
+            del self.rooms[pfrom][room]
+        else:
+            raise KeyError(
+                f'Unable to find the room {room} in the currently joined rooms'
+                + (f' for {pfrom}' if pfrom else '')
+            )
 
     def set_subject(self, room: JidStr, subject: str, *, mfrom: Optional[JID] = None):
         """Set a room’s subject.
@@ -538,10 +654,11 @@ class XEP_0045(BasePlugin):
 
     async def set_role(self, room: JidStr, nick: str, role: MucRole, *,
                        reason: str = '', ifrom: Optional[JidStr] = None, **iqkwargs):
-        """ Change role property of a nick in a room.
-            Typically, roles are temporary (they last only as long as you are in the
-            room), whereas affiliations are permanent (they last across groupchat
-            sessions).
+        """
+        Change role property of a nick in a room.
+        Typically, roles are temporary (they last only as long as you are in
+        the room), whereas affiliations are permanent (they last across
+        groupchat sessions).
 
         :param room: Room to modify.
         :param nick: User nickname to use in the set operation.
@@ -636,23 +753,69 @@ class XEP_0045(BasePlugin):
 
         :param room: Room to request voice from.
         """
-        #form = self.xmpp['xep_0004'].make_form(ftype='submit')
         msg = self.xmpp.make_message(room, mfrom=mfrom)
         form = msg['form']
         form['type'] = 'submit'
-        form.add_field(var='FORM_TYPE', ftype='hidden', value='http://jabber.org/protocol/muc#request')
-        form.add_field(var='muc#role', ftype='list-single', label='Requested role', value=role)
+        form.add_field(
+            var='FORM_TYPE',
+            ftype='hidden',
+            value='http://jabber.org/protocol/muc#request',
+        )
+        form.add_field(var='muc#role', ftype='list-single',
+                       label='Requested role', value=role)
         self.xmpp.send(msg)
 
-    def jid_in_room(self, room: JID, jid: JID) -> bool:
+    async def set_self_nick(self, room: JID, new_nick: str,
+                            timeout: int = 60,
+                            presence_options: Optional[PresenceArgs] = None) -> str:
+        """
+        Set your nickname in a room.
+        The room can arbitrarily decide on another nickname, so this function
+        waits for the response and returns the final nickname.
+
+        .. versionadded:: 1.9.0
+
+        :param room: Room in which to change our nickname
+        :param new_nick: Our new nickname in the room
+        :param timeout: Time to wait for our new nickname response
+        :param pfrom: (for components) the JID to send our presence from
+        """
+        new_jid = JID(room)
+        new_jid.resource = new_nick
+        if presence_options is None:
+            presence_options = {}
+        future = asyncio.Future()
+
+        def nickname_set(presence):
+            codes = presence['muc']['status_codes']
+            if 110 in codes and 303 in codes:
+                future.set_result(presence)
+
+        handler = self.xmpp.event_handler(
+            f"muc::{room}::self-presence",
+            nickname_set,
+        )
+        with handler:
+            self.xmpp.make_presence(pto=new_jid, **presence_options).send()
+            done, pending = await asyncio.wait([future], timeout=timeout)
+        if pending:
+            raise TimeoutError("Timed out waiting for server answer")
+
+        presence = future.result()
+        new_nick = presence['muc']['item_nick']
+        return new_nick
+
+    def jid_in_room(self, room: JID, jid: JID, pfrom: Optional[JID] = None) -> bool:
         """Check if a JID is present in a room.
 
         :param room: Room to check.
-        :param jid: FULL JID to check.
+        :param jid: full JID to check.
         """
         bare_match = False
-        for nick in self.rooms[room]:
-            entry = self.rooms[room][nick]
+        rooms = self.rooms.get(pfrom, {})
+        room = rooms.get(room, {})
+        for nick in room:
+            entry = room[nick]
             if not entry.get('jid'):
                 continue
 
@@ -660,7 +823,7 @@ class XEP_0045(BasePlugin):
                 return True
             elif JID(entry['jid']).bare == jid.bare:
                 bare_match = True
-        
+
         if bare_match:
             logging.info(
                 "Could not retrieve full JID, falling back to bare JID for %s in %s",
@@ -668,15 +831,17 @@ class XEP_0045(BasePlugin):
             )
         return bare_match
 
-    def get_nick(self, room: JID, jid: JID) -> Optional[str]:
+    def get_nick(self, room: JID, jid: JID, pfrom: Optional[JID] = None) -> Optional[str]:
         """Get the nickname of a specific JID in a room.
 
         :param room: Room to inspect.
         :param jid: FULL JID whose nick to return.
         """
         bare_match = None
-        for nick in self.rooms[room]:
-            entry = self.rooms[room][nick]
+        rooms = self.rooms.get(pfrom, {})
+        room = rooms.get(room, {})
+        for nick in room:
+            entry = room[nick]
             if not entry.get('jid'):
                 continue
 
@@ -684,7 +849,7 @@ class XEP_0045(BasePlugin):
                 return nick
             elif JID(entry['jid']).bare == jid.bare:
                 bare_match = nick
-        
+
         if bare_match:
             logging.info(
                 "Could not retrieve full JID, falling back to bare JID for %s in %s",
@@ -692,19 +857,20 @@ class XEP_0045(BasePlugin):
             )
         return bare_match
 
-    def get_joined_rooms(self) -> List[JID]:
+    def get_joined_rooms(self, pfrom: Optional[JID] = None) -> List[JID]:
         """Get the list of rooms we sent a join presence to
         and did not explicitly leave.
         """
-        return list(self.rooms.keys())
+        return list(self.rooms.get(pfrom, {}).keys())
 
-    def get_our_jid_in_room(self, room_jid: JID) -> str:
+    def get_our_jid_in_room(self, room_jid: JID, pfrom: Optional[JID] = None) -> str:
         """ Return the jid we're using in a room.
         """
-        return "%s/%s" % (room_jid, self.our_nicks[room_jid])
+        return "%s/%s" % (room_jid, self.our_nicks[pfrom][room_jid])
 
     def get_jid_property(self, room: JID, nick: str,
-                         jid_property: MucRoomItemKeys) -> Any:
+                         jid_property: MucRoomItemKeys,
+                         pfrom: Optional[JID] = None) -> Any:
         """ Get the property of a nick in a room, such as its 'jid' or 'affiliation'
             If not found, return None.
 
@@ -712,22 +878,28 @@ class XEP_0045(BasePlugin):
         :param nick: Which nickname information to get.
         :param jid_property: Property to fetch.
         """
-        if room in self.rooms and nick in self.rooms[room] and jid_property in self.rooms[room][nick]:
-            return self.rooms[room][nick][jid_property]
-        else:
-            return None
+        rooms = self.rooms.get(pfrom, {})
+        room_dict = rooms.get(room, {})
+        nick_dict = room_dict.get(nick, {})
+        prop = nick_dict.get(jid_property)
+        return prop or None
 
-    def get_roster(self, room: JID) -> List[str]:
+    def get_roster(self, room: JID, pfrom: Optional[JID] = None) -> List[str]:
         """ Get the list of nicks in a room.
 
         :param room: Room to list nicks from.
         """
-        if room not in self.rooms.keys():
+        rooms = self.rooms.get(pfrom, {})
+        if room not in rooms:
             raise ValueError("Room %s is not joined" % room)
-        return list(self.rooms[room].keys())
+        return list(rooms[room].keys())
 
     def get_users_by_affiliation(self, room: JidStr, affiliation='member', *, ifrom: Optional[JidStr] = None):
         # Preserve old API
         if affiliation not in AFFILIATIONS:
             raise ValueError("Affiliation %s does not exist" % affiliation)
         return self.get_affiliation_list(room, affiliation, ifrom=ifrom)
+
+    # Aliases muc→groupchat
+    join_groupchat = join_muc
+    join_groupchat_wait = join_muc_wait
