@@ -8,8 +8,12 @@ This file contains boilerplate to define types relevant to slixmpp.
 """
 
 from typing import (
+    Any,
+    Dict,
     Optional,
     Union,
+    Iterable,
+    List,
 )
 
 try:
@@ -36,6 +40,11 @@ PresenceShows = Literal[
     'away', 'chat', 'dnd', 'xa',
 ]
 
+# add the empty string, but not for sending
+ExtPresenceShows = Literal[
+    'away', 'chat', 'dnd', 'xa', ''
+]
+
 MessageTypes = Literal[
     'chat', 'error', 'groupchat',
     'headline', 'normal',
@@ -74,6 +83,39 @@ class MucRoomItem(TypedDict, total=False):
     alt_nick: str
 
 
+class ResourceDict(TypedDict, total=False):
+    show: ExtPresenceShows
+    priority: int
+    status: str
+
+
+RosterState = TypedDict(
+    'RosterState',
+    {
+        'from': bool,
+        'to': bool,
+        'pending_in': bool,
+        'pending_out': bool,
+        'whitelisted': bool,
+        'subscription': str,
+        'name': str,
+        'groups': List[str],
+        'removed': bool,
+    }
+)
+
+
+class RosterDBProtocol(Protocol):
+    def load(self, owner:JidStr, jid:JidStr, db_state: Dict[str, Any]) -> Optional[RosterState]:
+        ...
+
+    def save(self, owner:JidStr, jid:JidStr, state: RosterState, db_state: Dict[str, Any]):
+        ...
+
+    def entries(self, owner: OptJidStr, db_state: Optional[dict[str, Any]] = None) -> Iterable[str]:
+        ...
+
+
 MucRoomItemKeys = Literal[
     'jid', 'role', 'affiliation', 'show', 'status',  'alt_nick',
 ]
@@ -97,6 +139,7 @@ ErrorConditions = Literal[
     "not-allowed",
     "not-authorized",
     "payment-required",
+    "policy-violation",
     "recipient-unavailable",
     "redirect",
     "registration-required",
@@ -109,8 +152,21 @@ ErrorConditions = Literal[
     "unexpected-request",
 ]
 
+# https://xmpp.org/registrar/disco-categories.html#client
+ClientTypes = Literal[
+    "bot",
+    "console",
+    "game",
+    "handheld",
+    "pc",
+    "phone",
+    "sms",
+    "tablet",
+    "web",
+]
+
 __all__ = [
     'Protocol', 'TypedDict', 'Literal', 'OptJid', 'OptJidStr', 'JidStr', 'MAMDefault',
     'PresenceTypes', 'PresenceShows', 'MessageTypes', 'IqTypes', 'MucRole',
-    'MucAffiliation', 'FilterString', 'ErrorConditions', 'ErrorTypes'
+    'MucAffiliation', 'FilterString', 'ErrorConditions', 'ErrorTypes', 'ClientTypes'
 ]
